! 
!~ 30.4.2010: adapted from function_3p3m.f90 for internal masses
! 
!~ 24.6.2010: uses Andre van Hameren's OneLOop for finite C0
!
!****h* src/integral/three_point/function_3pC0i
! NAME
!
!  Module function_3p_finite
!
! USAGE
!
!  use function_3p_finite
!
! DESCRIPTION
!
!  This module is used to compute IR finite three point functions
!  with/without Feynman parameters in n, n+2 dimensions
!
! OUTPUT
!
!  This module exports the functions:
!  * f3p_finite,f3p_irfin_np2 -- functions for the computation of IR finite 
!  three-point functions with/without Feynman parameters in n, n+2 dimensions
!  * f3p_finite_c -- a function which computes the same thing as f3p_finite, only 
!    the format of the return values is different
!
! USES
!
!  * precision (src/module/precision_golem.f90)
!  * numerical_evaluation (src/numerical/mod_numeric.f90)
!  * dilogarithme (src/module/zdilog.f90)
!  * logarithme (src/module/z_log.f90)
!  * constante (src/module/constante.f90)
!  * parametre (src/module/parametre.f90)
!  * array (src/module/array.f90)
!  * sortie_erreur (src/module/sortie_erreur.f90)
!  * generic_function_2p (src/integrals/two_point/generic_function_2p.f90)
!  * multiply_div (src/module/multiply_div.f90)
!
!*****
module function_3p_finite
  !
  use precision_golem
  use numerical_evaluation
  use dilogarithme
  use logarithme
  use constante
  use parametre
  use array
  use equal
  use sortie_erreur
  use generic_function_2p
  use multiply_div
  implicit none
  !
  private :: ki
  real(ki) :: s12_glob,s23_glob,s13_glob
  real(ki) :: eps_glob
  integer :: par1_glob,par2_glob,par3_glob
  character (len=3) :: dim_glob
  !
  real(ki), dimension(3) :: b
  real(ki) :: sumb
  real(ki), dimension(3,3) :: invs,s_mat
  integer, dimension(3) :: par
  integer, dimension(3) :: s = (/1,2,3/)
  !
  logical, dimension(:), allocatable :: deja_calcule
  real(ki),dimension(:,:), allocatable :: resultat
  logical, dimension(:,:), allocatable :: deja_calcule2
  real(ki),dimension(:,:,:), allocatable :: resultat2
  logical, dimension(:), allocatable :: deja_calcule_np2
  real(ki),dimension(:,:), allocatable :: resultat_np2
  logical, dimension(:,:,:), allocatable :: deja_calcule22
  real(ki),dimension(:,:,:,:), allocatable :: resultat22
  !
 private :: eps_glob,s12_glob,s23_glob,s13_glob,par1_glob,par2_glob,par3_glob,dim_glob
 private :: b,sumb,invs,s_mat,par,s
 private :: deja_calcule,resultat,deja_calcule2,resultat2,deja_calcule_np2,resultat_np2,deja_calcule22,resultat22
  public :: f3p_finite,f3p_finite_c,C0
  !
  !
      interface
        subroutine avh_olo_c0m(rslt,p1,p2,p3,m1,m2,m3)
           use precision_golem, only: ki_avh
           implicit none
           complex(ki_avh), intent(out) :: rslt(0:2)
           real(ki_avh), intent(in) :: p1,p2,p3,m1,m2,m3
        end subroutine avh_olo_c0m
      end interface
       !
  contains
    !
    !****f* src/integral/three_point/function_3p_finite/f3p_finite
    ! NAME
    !
    !  Function f3p_finite
    !
    ! USAGE
    !
    !  real_dim4 = f3p_finite(dim,s1,s2,s3,m1,m2,m3,par1,par2,par3)
    !
    ! DESCRIPTION
    !
    !  This function computes the IR finite three-point 
    !  function in n and n+2 dimensions. 
    !
    ! INPUTS
    !
    !  * s1 -- a real (type ki), p1^2
    !  * s2 -- a real (type ki), p2^2
    !  * s3 -- a real (type ki), p3^2
    !  * m1 -- a real (type ki), the first mass squared
    !  * m2 -- a real (type ki), the second mass squared
    !  * m3 -- a real (type ki), the third mass squared
    !  * par1 -- an integer, the label of the third Feynman parameter
    !  * par2 -- an integer, the label of the second Feynman parameter
    !  * par3 -- an integer, the label of the first Feynman parameter
    !  Note that par1,par2 and par3 are supposed to be ordered, i.e.
    !  par1 <= par2 <= par3, note also that put zero for par1, par2 or par3
    !  if this Feynman parameter does not exist.
    !  Use the routine tri_int(t_in,t_out) to order the labels in the module 
    !  tri_croissant (src/module/tri.f90)
    !
    ! SIDE EFFECTS
    !
    !  No side effect
    !
    ! RETURN VALUE
    !
    !  An real (type ki) array of rank 1 and shape 6,
    !  where the last two entries corresponding to 
    !  the real/imaginary part of the constant term. 
    !  the first 4 entries are always zero, but the shape should be 
    !  uniform for all triangles called in generic_function_3p
    !
    !*****
    function f3p_finite(dim,s1,s2,s3,m1,m2,m3,par1,par2,par3)
      implicit none
      !
      character (len=3), intent (in) :: dim
      real(ki), intent (in) :: s1,s2,s3,m1,m2,m3
      integer, intent (in) :: par1,par2,par3
      real(ki), dimension(4) :: f3p_finite
      integer :: nb_par
      real(ki) :: lamb,detS3
      real(ki) :: plus_grand
      real(ki) :: norma
      complex(ki) :: resto,abserro
      !
      par = (/par1,par2,par3/)
      !
      s_mat(1,:) = (/-m1*2._ki,s2-m1-m2,s1-m1-m3/)
      s_mat(2,:) = (/s2-m1-m2,-m2*2._ki,s3-m2-m3/)
      s_mat(3,:) = (/s1-m1-m3,s3-m2-m3,-m3*2._ki/)
      !
  ! on redefinit la matrice S de telle facon a ce que ces elements
      ! soient entre -1 et 1
      ! this can't be done if partly invariants instead of 
      ! s_mat are used as arguments
      !
           plus_grand = 1._ki
    ! if (rat_or_tot_par%tot_selected) then
        !
       !  plus_grand = maxval(array=abs(s_mat))
      !
      ! else if (rat_or_tot_par%rat_selected) then
       !
      !   plus_grand = 1._ki
       !
      ! end if
      !
       s_mat = s_mat/plus_grand
      !
      detS3 = -(s_mat(1, 3)**2*s_mat(2, 2)) + 2*s_mat(1, 2)*s_mat(1, 3)*s_mat(2, 3) -  &
     &     s_mat(1, 2)**2*s_mat(3, 3) + s_mat(1, 1)*(-s_mat(2, 3)**2 +   &
     &     s_mat(2, 2)*s_mat(3, 3))
      
      b(1) = (-s_mat(2, 3)**2 + s_mat(1, 3)*(-s_mat(2, 2) + s_mat(2, 3)) +     &
     &      s_mat(1, 2)*(s_mat(2, 3) - s_mat(3, 3)) + s_mat(2, 2)*s_mat(3, 3))/detS3
      b(2) = (-s_mat(1, 3)**2 + s_mat(1, 3)*s_mat(2, 3) +  &
     &    s_mat(1, 2)*(s_mat(1, 3) - s_mat(3, 3)) +   &
     &    s_mat(1, 1)*(-s_mat(2, 3) + s_mat(3, 3)))/detS3
      b(3) = (-s_mat(1, 2)**2 - s_mat(1, 3)*s_mat(2, 2) +   &
     &    s_mat(1, 1)*(s_mat(2, 2) - s_mat(2, 3)) +      &
     &    s_mat(1, 2)*(s_mat(1, 3) + s_mat(2, 3)))/detS3
      !
      sumb = (-s_mat(1, 2)**2 - s_mat(1, 3)**2 + s_mat(1, 1)*s_mat(2, 2) - &
     &   2*s_mat(1, 3)*(s_mat(2, 2) - s_mat(2, 3)) - 2*s_mat(1, 1)*s_mat(2, 3) - &
     &   s_mat(2, 3)**2 + 2*s_mat(1, 2)*(s_mat(1, 3) + s_mat(2, 3) - s_mat(3, 3)) + &
     &   s_mat(1, 1)*s_mat(3, 3) + s_mat(2, 2)*s_mat(3, 3))/detS3
      !
      invs(1,1) = (-s_mat(2, 3)**2 + s_mat(2, 2)*s_mat(3, 3))/detS3
      invs(1,2) = (s_mat(1, 3)*s_mat(2, 3) - s_mat(1, 2)*s_mat(3, 3))/detS3
      invs(1,3) = (-(s_mat(1, 3)*s_mat(2, 2)) + s_mat(1, 2)*s_mat(2, 3))/detS3
      invs(2,1) = invs(1,2)
      invs(2,2) = (-s_mat(1, 3)**2 + s_mat(1, 1)*s_mat(3, 3))/detS3
      invs(2,3) = (s_mat(1, 2)*s_mat(1, 3) - s_mat(1, 1)*s_mat(2, 3))/detS3
      invs(3,1) = invs(1,3)
      invs(3,2) = invs(2,3)
      invs(3,3) = (-s_mat(1, 2)**2 + s_mat(1, 1)*s_mat(2, 2))/detS3
      !
      lamb = 2._ki*s_mat(1,3)*s_mat(2,3)+2._ki*s_mat(1,2)*s_mat(2,3)&
      +2._ki*s_mat(1,3)*s_mat(1,2)-s_mat(1,3)*s_mat(1,3)-s_mat(1,2)*s_mat(1,2)&
      -s_mat(2,3)*s_mat(2,3)
      !
         nb_par = count(mask=par/=0)
      !
      if (nb_par == 0) then
        !
        norma = -1._ki/2._ki
        !
      else if (nb_par == 1) then
        !
        norma = -1._ki/6._ki
        !
      else
        !
        norma = 0._ki
        !
      end if
      !
      ! memory allocation to save time in the recursion
      !
      allocate(deja_calcule(4))
      allocate(resultat(4,2))
      allocate(deja_calcule2(3,4))
      allocate(resultat2(3,4,4))
      allocate(deja_calcule_np2(4))
      allocate(resultat_np2(4,4))
      allocate(deja_calcule22(3,4,4))
      allocate(resultat22(3,4,4,4))
      !
      ! initialisation
      !
      deja_calcule = .false.
      resultat = 0._ki
      deja_calcule2 = .false.
      resultat2 = 0._ki
      deja_calcule_np2 = .false.
      resultat_np2 = 0._ki
      deja_calcule22 = .false.
      resultat22 = 0._ki
      !
      !
      f3p_finite = 0._ki
      !
      !      if (abs(sumb) > coupure_3p3m) then
      if (abs(sumb) > 0._ki) then
        !
        ! always use analytic computation until massive numerical is implemented
	! when invariants as input are used, no division by plus_grand!
        !
        if (dim == "ndi") then
          !
          f3p_finite(3:4)= a3pC0i(s1,s2,s3,m1,m2,m3,par1,par2,par3)
          !
        else if (dim == "n+2") then
          !
          f3p_finite = a3pC0i_np2(s1,s2,s3,m1,m2,m3,par1,par2,par3)
         ! f3p_finite(3) = f3p_finite(3)-log(plus_grand)*norma
	  ! mu2_scale_par is already contained in the bubbles, 
	  ! scaling of s_mat  does not enter here
          !
        end if
	!
      else
        !
        ! numerical computation
        !
        dim_glob = dim
        par1_glob = par1
        par2_glob = par2
        par3_glob = par3
        !
        s13_glob = s_mat(1,3)
        s12_glob = s_mat(1,2)
        s23_glob = s_mat(2,3)
        !
        resto = 0._ki
        abserro = 0._ki
        !
        ! on pose z = x - i*eps*y (avec x et y > 0)
        ! z*s13+(1-z)*s23 = s23+x*(s13-s23)-i*eps*y*(s13-s23)
        ! on veut la partie imaginaire du meme signe que i*lambda
        ! => eps*(s13-s23) < 0
        !
        ! faire attention que suivant le signe de eps_glob, on tourne dans le
        ! sens des aiguilles d'une montre ou inversement
        ! eps_glob = 1, on ferme le contour vers le bas --> -2 i Pi residu
        ! eps_glob = -1, on ferme le contour vers le haut --> +2 i Pi residu
        !
        eps_glob = sign(1._ki,s23_glob-s13_glob)
        !
        origine_info_par = "f3p_finite, dimension "//dim
        num_grand_b_info_par = lamb
        denom_grand_b_info_par = (2._ki*s_mat(1,3)*s_mat(1,2)*s_mat(2,3))
        !
        !
        if (dim == "ndi") then      
          !
          resto = resto/plus_grand
          !
        else if (dim == "n+2") then
          !
          f3p_finite(1) = norma
          f3p_finite(2) = 0._ki
          resto = resto-log(plus_grand/mu2_scale_par)*norma
          !
        end if
        !
        f3p_finite(3) = real(resto,ki)
        f3p_finite(4) = aimag(resto)
        !
      end if  ! end if analytic or numeric
      !
      ! on libere la memoire
      !
      deallocate(deja_calcule)
      deallocate(resultat)
      deallocate(deja_calcule2)
      deallocate(resultat2)
      deallocate(deja_calcule_np2)
      deallocate(resultat_np2)
      deallocate(deja_calcule22)
      deallocate(resultat22)
      !
    end function f3p_finite
    !
    !****f* src/integral/three_point/function_3pC0i/f3p_finite_c
    ! NAME
    !
    !  Function f3p_finite_c
    !
    ! USAGE
    !
    !  complex_dim2 = f3p_finite_c(s1,s2,s3,m1,m2,m3,par1,par2,par3)
    !
    ! DESCRIPTION
    !
    !  It computes the same as the function f3p_finite, but the returned
    !  value is a complex (type ki) array of rank 1 and shape 2
    !
    ! INPUTS
    !
    !  * dim -- a character (length 3), to compute in n or n+2 dimensions
    !  * s1 -- a real (type ki), p1^2
    !  * s2 -- a real (type ki), p2^2
    !  * s3 -- a real (type ki), p3^2
    !  * m1 -- a real (type ki), the first mass^2
    !  * m2 -- a real (type ki), the second mass^2
    !  * m3 -- a real (type ki), the third mass^2
    !  * par1 -- an integer, the label of the third Feynman parameter
    !  * par2 -- an integer, the label of the second Feynman parameter
    !  * par3 -- an integer, the label of the first Feynman parameter
    !  Note that par1,par2 and par3 are supposed to be ordered, i.e.
    !  par1 <= par2 <= par3, note also that put zero for par1, par2 or par3
    !  if this Feynman parameter does not exist.
    !  Use the routine tri_int(t_in,t_out) to order the labels in the module 
    !  tri_croissant (src/module/tri.f90)
    !
    ! SIDE EFFECTS
    !
    !  No side effect
    !
    ! RETURN VALUE
    !
    !  An complex (type ki) array of rank 1 and shape 2 corresponding to 
    !  the (real part,imaginary part) of the coefficient of the 1/epsilon term
    !  and the (real part,imaginary part) of the constant term. if par1 and/or par2
    !  are different from zero for dim="n+2", an error is returned.
    !
    ! EXAMPLE
    !
    !
    !
    !*****
    function f3p_finite_c(dim,s1,s2,s3,m1,m2,m3,par1,par2,par3)
      !
      use translate
      implicit none
      !
      character (len=3), intent (in) :: dim
      real(ki), intent (in) :: s1,s2,s3,m1,m2,m3
      integer, intent (in) :: par1,par2,par3
      complex(ki), dimension(2) :: f3p_finite_c
      !
      real(ki), dimension(4) :: res4
      !
      res4 = f3p_finite(dim,s1,s2,s3,m1,m2,m3,par1,par2,par3)
      call to_complex(res4,f3p_finite_c)
      !
    end function f3p_finite_c
    !
    !****if* src/integral/three_point/function_3pC0i/a3pC0i
    ! NAME
    !
    !  Function a3pC0i
    !
    ! USAGE
    !
    !  real_dim2 = a3pC0i(s1,s2,s3,m1,m2,m3,par1,par2,par3)
    !
    ! DESCRIPTION
    !
    !  This recursive function implements the formula of ref 1
    !
    ! INPUTS
    !
    !  * s1 -- a real (type ki), p1^2
    !  * s2 -- a real (type ki), p2^2
    !  * s3 -- a real (type ki), p3^2
    !  * m1 -- a real (type ki), the first internal mass squared
    !  * m2 -- a real (type ki), the second internal mass squared
    !  * m3 -- a real (type ki), the third internal mass squared
    !  * par1 -- an integer, the label of the third Feynman parameter
    !  * par2 -- an integer, the label of the second Feynman parameter
    !  * par3 -- an integer, the label of the first Feynman parameter
    !
    ! SIDE EFFECTS
    !
    !  This function modifies the value of the local (for the module) variables:
    !  * deja_calcule, deja_calcule2, deja_calcule_np2 and deja_calcule22
    !  * resultat, resultat2, resultat_np2 and resultat22
    !
    ! RETURN VALUE
    !
    !  It returns a real (type ki) array of rank 1 and shape 2
    !
    ! EXAMPLE
    !
    !
    !
    !*****
    recursive function a3pC0i(s1,s2,s3,m1,m2,m3,par1,par2,par3) result(res_3pC0i)
      !
      implicit none
      real(ki), intent (in) :: s1,s2,s3,m1,m2,m3
      integer, intent (in) :: par1,par2,par3
      real(ki), dimension(2) :: res_3pC0i
      !
      integer :: j
      integer :: nb_par_loc
      integer, dimension(3) :: par_loc,par_plus
      real(ki), dimension(4) :: truc1
      real(ki), dimension(2) :: temp0
      real(ki), dimension(4) :: temp1,temp2,temp3
      real(ki), dimension(4) :: temp10,temp11,temp12
      complex(ki) :: ctemp
      integer :: ib,b_pro,b_pro_mj
      !
      b_pro = packb(s)
      !
      par_loc = (/par1,par2,par3/)
      par_plus = par_loc+1
      nb_par_loc = count(mask=par_loc/=0)
      !
      ! cas sans parametre de feynman au numerateur
      !
      if (nb_par_loc == 0) then
        ctemp = C0(s1,s2,s3,m1,m2,m3)
        res_3pC0i(1) = real(ctemp,ki)
        res_3pC0i(2) = aimag(ctemp)
      !
      ! cas avec un parametre de feynman au numerateur
      !
      else if (nb_par_loc == 1) then
        !
        if (deja_calcule(1)) then
          !
          temp0 = resultat(1,:)
          !
        else
          !
          temp0 = a3pC0i(s1,s2,s3,m1,m2,m3,0,0,0)
          resultat(1,:) = temp0
          deja_calcule(1) = .true.
          !
        end if
        !
        temp1 = 0._ki
        temp2 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if (deja_calcule2(j,1)) then
              !
              truc1 = resultat2(j,1,:)
              !
            else
              !
              truc1 = f2p(s_mat,b_pro_mj)
              resultat2(j,1,:) = truc1
              deja_calcule2(j,1) = .true.
              !
            end if
            !
            temp1 = temp1 + b(j)*truc1
            temp2 = temp2 + invs(j,par3)*truc1
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        res_3pC0i(1) = b(par3)*(temp0(1) - temp1(3))/sumb + temp2(3)
        res_3pC0i(2) = b(par3)*(temp0(2) - temp1(4))/sumb + temp2(4)
      !
      ! cas avec deux parametres de feynman au numerateur
      !
      else if (nb_par_loc == 2) then
        !
        if (deja_calcule_np2(par_plus(3))) then
          !
          temp11 = resultat_np2(par_plus(3),:)
          !
        else
          !
          temp11 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,0,0,par3)
          resultat_np2(par_plus(3),:) = temp11
          deja_calcule_np2(par_plus(3)) = .true.
          !
        end if
        !
        temp10 = resultat_np2(1,:)
        temp3 = invs(par2,par3)*temp10
        temp1 = b(par2)*temp11
        temp1 = mult_div(-2._ki/3._ki,temp1)*3._ki
        temp2 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if (j /= par3) then
              !
              if (deja_calcule2(j,par_plus(3))) then
                !
                truc1 = resultat2(j,par_plus(3),:)
                !
              else
                !
                truc1 = f2p(s_mat,b_pro_mj,par3)
                resultat2(j,par_plus(3),:) = truc1
                deja_calcule2(j,par_plus(3)) = .true.
                !
              end if
              !
              temp2 = temp2 + invs(j,par2)*truc1
              !
            end if
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        res_3pC0i(1) = -temp3(3) + temp1(3) + temp2(3)
        res_3pC0i(2) = -temp3(4) + temp1(4) + temp2(4)
      !
      ! cas avec trois parametres de feynman au numerateur
      !
      else
        !
        temp12 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,0,par2,par3)
        temp10 = resultat_np2(par_plus(3),:)
        temp11 = resultat_np2(par_plus(2),:)
        temp3 = invs(par1,par2)*temp10 &
                + invs(par1,par3)*temp11
        temp1 = b(par1)*temp12
        temp1 = mult_div(-1._ki/2._ki,temp1)*4._ki
        temp2 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if ( (j /= par3) .and. (j /= par2) ) then
              !
              truc1 = resultat22(j,par_plus(2),par_plus(3),:)
              temp2 = temp2 + invs(j,par1)*truc1
              !
            end if
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        res_3pC0i(1) = -temp3(3) + temp1(3) + temp2(3)
        res_3pC0i(2) = -temp3(4) + temp1(4) + temp2(4)
        !
      end if
      !
    end function a3pC0i
    !
    !****if* src/integral/three_point/function_3pC0i/a3pC0i_np2
    ! NAME
    !
    !  Function a3pC0i_np2
    !
    ! USAGE
    !
    !  real_dim4 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,par1,par2,par3)
    !
    ! DESCRIPTION
    !
    !  This recursive function implements the formula of ref 1
    !
    ! INPUTS
    !
    !  * s1 -- a real (type ki), p1^2
    !  * s2 -- a real (type ki), p2^2
    !  * s3 -- a real (type ki), p3^2
    !  * m1 -- a real (type ki), the first internal mass squared
    !  * m2 -- a real (type ki), the second internal mass squared
    !  * m3 -- a real (type ki), the third internal mass squared
    !  * par1 -- an integer, the label of the third Feynman parameter
    !  * par2 -- an integer, the label of the second Feynman parameter
    !  * par3 -- an integer, the label of the first Feynman parameter
    !
    ! SIDE EFFECTS
    !
    !  This function modifies the value of the local (for the module) variables:
    !  * deja_calcule, deja_calcule2, deja_calcule_np2 and deja_calcule22
    !  * resultat, resultat2, resultat_np2 and resultat22
    !
    ! RETURN VALUE
    !
    !  It returns a real (type ki) array of rank 1 and shape 4
    !
    ! EXAMPLE
    !
    !
    !
    !*****
    recursive function a3pC0i_np2(s1,s2,s3,m1,m2,m3,par1,par2,par3) result(res_3pC0i_np2)
      !
      implicit none

      real(ki), intent (in) :: s1,s2,s3,m1,m2,m3
      integer, intent (in) :: par1,par2,par3
      real(ki), dimension(4) :: res_3pC0i_np2
      !
      integer :: j
      integer :: nb_par_loc
      integer, dimension(3) :: par_loc,par_plus
      real(ki), dimension(4) :: truc1,truc2
      real(ki), dimension(2) :: temp0
      real(ki), dimension(4) :: temp1,temp2,temp3
      real(ki), dimension(4) :: temp10,temp11
      integer :: ib,b_pro,b_pro_mj
      !
      b_pro = packb(s)
      !
      par_loc = (/par1,par2,par3/)
      par_plus = par_loc+1
      nb_par_loc = count(mask=par_loc/=0)
      !
      ! cas sans parametre de feynman au numerateur
      !
      if (nb_par_loc == 0) then
        !
        if (deja_calcule(1)) then
          !
          temp0 = resultat(1,:)
          !
        else
          !
          temp0 = a3pC0i(s1,s2,s3,m1,m2,m3,0,0,0)
          resultat(1,:) = temp0
          deja_calcule(1) = .true.
          !
        end if
        !
        temp1 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if (deja_calcule2(j,1)) then
              !
              truc1 = resultat2(j,1,:)
              !
            else
              !
              truc1 = f2p(s_mat,b_pro_mj)
              resultat2(j,1,:) = truc1
              deja_calcule2(j,1) = .true.
              !
            end if
            !
            temp1 = temp1 + b(j)*truc1
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        res_3pC0i_np2(1) = (- temp1(1))/sumb
        res_3pC0i_np2(2) = (- temp1(2))/sumb
        res_3pC0i_np2(3) = (temp0(1) - temp1(3))/sumb
        res_3pC0i_np2(4) = (temp0(2) - temp1(4))/sumb
        res_3pC0i_np2 = mult_div(1._ki,res_3pC0i_np2)/2._ki
      !
      ! cas avec un parametre de feynman au numerateur
      !
      else if (nb_par_loc == 1) then
        !
        if (deja_calcule_np2(1)) then
          !
          temp10 = resultat_np2(1,:)
          !
        else
          !
          temp10 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,0,0,0)
          resultat_np2(1,:) = temp10
          deja_calcule_np2(1) = .true.
          !
        end if
        !
        temp3 = b(par3)*temp10
        temp1 = 0._ki
        temp2 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if (deja_calcule2(j,1)) then
              !
              truc1 = resultat2(j,1,:)
              !
            else
              !
              truc1 = f2p(s_mat,b_pro_mj)
              resultat2(j,1,:) = truc1
              deja_calcule2(j,1) = .true.
              !
            end if
            !
            temp1 = temp1 + invs(j,par3)*truc1
            !
            if (j /= par3) then
              !
              if (deja_calcule2(j,par_plus(3))) then
                !
                truc2 = resultat2(j,par_plus(3),:)
                !
              else
                !
                truc2 = f2p(s_mat,b_pro_mj,par3)
                resultat2(j,par_plus(3),:) = truc2
                deja_calcule2(j,par_plus(3)) = .true.
                !
              end if
              !
              temp2 = temp2 + b(j)*truc2
              !
            end if
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        temp1 = mult_div(2._ki/3._ki,temp1)/3._ki
        temp2 = mult_div(2._ki/3._ki,temp2)/3._ki
        res_3pC0i_np2(1) = (temp3(1) + temp1(1) - temp2(1))/sumb
        res_3pC0i_np2(2) = (temp3(2) + temp1(2) - temp2(2))/sumb
        res_3pC0i_np2(3) = (temp3(3) + temp1(3) - temp2(3))/sumb
        res_3pC0i_np2(4) = (temp3(4) + temp1(4) - temp2(4))/sumb
      !
      ! cas avec deux parametres de feynman au numerateur
      !
      else if (nb_par_loc == 2) then
        !
        temp0 = a3pC0i(s1,s2,s3,m1,m2,m3,0,par2,par3)
        !
        if (deja_calcule_np2(par_plus(2))) then
          !
          temp10 = resultat_np2(par_plus(2),:)
          !
        else
          !
          temp10 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,0,0,par2)
          resultat_np2(par_plus(2),:) = temp10
          deja_calcule_np2(par_plus(2)) = .true.
          !
        end if
        !
        if (deja_calcule_np2(par_plus(3))) then
          !
          temp11 = resultat_np2(par_plus(3),:)
          !
        else
          !
          temp11 = a3pC0i_np2(s1,s2,s3,m1,m2,m3,0,0,par3)
          resultat_np2(par_plus(3),:) = temp11
          deja_calcule_np2(par_plus(3)) = .true.
          !
        end if
        !
        temp3 =  b(par3)*temp10 + b(par2)*temp11
        temp1 = 0._ki
        !
        ib = b_pro
        j = 0
        !
        do while (ib /= 0)
          !
          if (modulo(ib,2) == 1)  then
            !
            b_pro_mj = ibclr(b_pro,j)
            !
            if ( (j /= par2) .and. (j /= par3) ) then
              !
              if (deja_calcule22(j,par_plus(2),par_plus(3))) then
                !
                truc1 = resultat22(j,par_plus(2),par_plus(3),:)
                !
              else
                !
                truc1 = f2p(s_mat,b_pro_mj,par2,par3)
                resultat22(j,par_plus(2),par_plus(3),:) = truc1
                deja_calcule22(j,par_plus(2),par_plus(3)) = .true.
                !
              end if
              !
              temp1 = temp1 + b(j)*truc1
              !
            end if
            !
          end if
          !
          j = j+1
          ib= ishft(ib,-1)
          !
        end do
        !
        res_3pC0i_np2(1) = (temp3(1) - temp1(1))/sumb
        res_3pC0i_np2(2) = (temp3(2) - temp1(2))/sumb
        res_3pC0i_np2(3) = (temp0(1) + temp3(3) - temp1(3))/sumb
        res_3pC0i_np2(4) = (temp0(2) + temp3(4) - temp1(4))/sumb
        res_3pC0i_np2 = mult_div(1._ki/2._ki,res_3pC0i_np2)/4._ki
      !
      ! cas avec trois parametres de feynman au numerateur
      !
      else
        !
        tab_erreur_par(1)%a_imprimer = .true.
        tab_erreur_par(1)%chaine = 'Error in f3p_finite:'
        tab_erreur_par(2)%a_imprimer = .true.
        tab_erreur_par(2)%chaine = 'rank 3 6-dim 3-point function should not be needed'
        call catch_exception(0)
        !
      end if
      !
    end function a3pC0i_np2
    !
    !****f* src/integral/three_point/function_3pC0i/C0
    ! NAME
    !
    ! Function C0, follows A.Denner, Fortsch.Phys.41:307-420,1993
    ! resp. Oldenborgh/Vermaseren Z.Phys.C46:425-438,1990
    !
    ! USAGE
    !
    !  complex = C0(s1,s2,s3,m1,m2,m3)
    !
    ! DESCRIPTION
    !
    !  This function computes finite scalar three point functions
    !  with internal masses in 4 dimensions
    !
    ! INPUTS
    !
    !  * s1 -- a real (type ki), p1^2
    !  * s2 -- a real (type ki), p2^2
    !  * s3 -- a real (type ki), p3^2
    !  * m1 -- a real (type ki), the first internal mass squared
    !  * m2 -- a real (type ki), the second internal mass squared
    !  * m3 -- a real (type ki), the third internal mass squared
    !
    ! SIDE EFFECTS
    !
    !  No side effect, it uses the value of rat_or_tot_par 
    !  (in src/module/parametre.f90)
    !
    ! RETURN VALUE
    !
    !  It returns a complex (type ki)
    !
    !  
    !*****
      function C0(s1,s2,s3,m1sq,m2sq,m3sq)
      implicit none
!
!  s1,s2,s3 = SQUARED external momenta
!  m1sq,m2sq,m3sq = SQUARED internal masses
      real(ki), intent(in) :: s1,s2,s3,m1sq,m2sq,m3sq
      real(ki) :: del
      complex(ki) :: C0
      complex(ki_avh), dimension(0:2) :: C0olo
      integer :: i,j,k
      !  del = epsilon(1._ki)
      !
      if (rat_or_tot_par%tot_selected) then
      !
         if (.not. olo) then
       !    call avh_olo_onshell(100._ki*del)
          call avh_olo_onshell(1.e-10_ki)
          call avh_olo_mu_set(sqrt(mu2_scale_par))
          olo=.true.
         end if
         call avh_olo_c0m(C0olo,real(s1,ki_avh),real(s2,ki_avh),real(s3,ki_avh), &
                    &    real(m3sq,ki_avh),real(m1sq,ki_avh),real(m2sq,ki_avh))
    !
         C0=C0olo(0)
    !
       else if (rat_or_tot_par%rat_selected) then
        !
         C0 = 0._ki
        !
       end if
      !
      end function C0
    !
    ! *************************************************
    !
end module function_3p_finite
!
